@@ -4,10 +4,8 @@ import itertools
 import json
 import os
 import re
-import shutil
 import uuid
 from enum import Enum
-from pathlib import Path
 from typing import List, Optional, Any
 
 import msgpack
@@ -50,6 +48,22 @@ class RemoteStore:
         """
         pass
 
+    def delete_contents(self):
+        """
+        Delete the contents of the store
+        """
+        from kodexa import KodexaPlatform
+        import requests
+        resp = requests.delete(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.get_ref().replace(':','/')}/contents",
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+
+        if resp.status_code == 200:
+            return resp.content
+        else:
+            msg = f"Unable to delete contents {resp.text}, status : {resp.status_code}"
+            raise Exception(msg)
+
 
 class DocumentMetadata(Dict):
     """
@@ -62,12 +76,15 @@ class DocumentMetadata(Dict):
 
 class Tag(Dict):
 
-    def __init__(self, start=None, end=None, value=None, data=None, *args, **kwargs):
+    def __init__(self, start: Optional[int] = None, end: Optional[int] = None, value: Optional[str] = None,
+                 uuid: Optional[str] = None, data: Any = None, *args,
+                 **kwargs):
         super().__init__(*args, **kwargs)
-        self.start = start
-        self.end = end
-        self.value = value
-        self.data = data
+        self.start: Optional[int] = start
+        self.end: Optional[int] = end
+        self.value: Optional[str] = value
+        self.data: Optional[Any] = data
+        self.uuid: Optional[str] = uuid
 
 
 class FindDirection(Enum):
@@ -102,12 +119,12 @@ class ContentNode(object):
         <kodexa.model.model.ContentNode object at 0x7f80605e53c8>
         >>> current_content_node.add_child(new_page)
     """
-
-    def __init__(self, document, node_type: str, content="", content_parts=None):
+    def __init__(self, document, node_type: str, content: Optional[str] = None,
+                 content_parts: Optional[List[Any]] = None):
         if content_parts is None:
             content_parts = []
         self.node_type: str = node_type
-        self.content: str = content
+        self.content: Optional[str] = content
         self.document: Document = document
         self.content_parts: List[Any] = content_parts
         self.parent: Optional[ContentNode] = None
@@ -544,6 +561,43 @@ class ContentNode(object):
         """
         self.remove_feature('tag', tag_name)
 
+    def copy_tag(self, selector=".", existing_tag_name=None, new_tag_name=None):
+        """
+        Creates a new tag of 'new_tag_name' on the selected content node(s) with the same information as the tag with 'existing_tag_name'.
+        Both existing_tag_name and new_tag_name values are required and must be different from one another.  Otherwise, no action is taken.
+        If a tag with the 'existing_tag_name' does not exist on a selected node, no action is taken for that node.
+
+            >>> document.get_root().copy_tag('foo', 'bar')
+
+        :param selector: The selector to identify the source nodes to work on (default . - the current node)
+        :param str existing_tag_name: The name of the existing tag whose values will be copied to the new tag.
+        :param str new_tag_name: The name of the new tag.  This must be different from the existing_tag_name.
+        """
+        if existing_tag_name == None or new_tag_name == None or existing_tag_name == new_tag_name:
+            return # do nothing, just exit function
+
+        for node in self.select(selector):
+            existing_tag_values = node.get_feature_value('tag', existing_tag_name)
+            if existing_tag_values:
+                if type(existing_tag_values) == list:
+
+                    # It's possible to have multiple features with the same tag name that also share the same uuid.
+                    # If we DO have features with the same UUID, we need to make sure that their copies also share the same UUID.
+                    sorted_tag_values = sorted(existing_tag_values, key=lambda k: k['uuid']) 
+                    previous_uuid = None
+                    new_uuid = None
+                    for val in sorted_tag_values:
+                        if previous_uuid == None or previous_uuid != val['uuid']:
+                            new_uuid = str(uuid.uuid4())
+                        
+                        previous_uuid = val['uuid']
+                        tag = Tag(start=val['start'], end=val['end'], value=val['value'], uuid=new_uuid, data=val['data'])
+                        node.add_feature('tag', new_tag_name, tag)
+                else:
+                    tag = Tag(start=existing_tag_values['start'], end=existing_tag_values['end'], value=existing_tag_values['value'], uuid=str(uuid.uuid4()), data=existing_tag_values['data'])
+                    node.add_feature('tag', new_tag_name, tag)
+
+        
     def collect_nodes_to(self, end_node):
         """
         Get the the sibling nodes between the current node and the end_node.
@@ -565,7 +619,7 @@ class ContentNode(object):
                 break
         return nodes
 
-    def tag_nodes_to(self, end_node, tag_to_apply):
+    def tag_nodes_to(self, end_node, tag_to_apply, tag_uuid: str = None):
         """
         Tag all the nodes from this node to the end_node with the given tag name
 
@@ -573,8 +627,9 @@ class ContentNode(object):
 
         :param ContentNode end_node: The node to end with
         :param str tag_to_apply: The tag name that will be applied to each node
+        :param str tag_uuid: The tag uuid used if you want to group them
         """
-        [node.tag(tag_to_apply) for node in self.collect_nodes_to(end_node)]
+        [node.tag(tag_to_apply, tag_uuid=tag_uuid) for node in self.collect_nodes_to(end_node)]
 
     def tag_range(self, start_content_re, end_content_re, tag_to_apply, node_type_re='.*', use_all_content=False):
         """
@@ -623,7 +678,7 @@ class ContentNode(object):
 
     def tag(self, tag_to_apply, selector=".", content_re=None,
             use_all_content=False, node_only=None,
-            fixed_position=None, data=None, separator=" "):
+            fixed_position=None, data=None, separator=" ", tag_uuid: str = None):
         """
         This will tag (see Feature Tagging) the expression groups identified by the regular expression.
 
@@ -632,15 +687,22 @@ class ContentNode(object):
         Note that if you use the flag use_all_content then node_only will default to True if not set, else it
         will default to False
 
-        :param tag_to_apply: the name of tag that will be applied to the node
+        :param tag_to_apply: The name of tag that will be applied to the node
         :param selector: The selector to identify the source nodes to work on (default . - the current node)
-        :param content_re: the regular expression that you wish to use to tag, note that we will create a tag for each matching group
-        :param use_all_content: apply the regular expression to the all_content (include content from child nodes)
+        :param content_re: The regular expression that you wish to use to tag, note that we will create a tag for each matching group
+        :param use_all_content: Apply the regular expression to the all_content (include content from child nodes)
         :param separator: Separator to use for use_all_content
         :param node_only: Ignore the matching groups and tag the whole node
-        :param fixed_position: use a fixed position, supplied as a tuple i.e. - (4,10) tag from position 4 to 10 (default None)
-        :param data: Attach the a dictionary of data for the given tag
-
+        :param fixed_position: Use a fixed position, supplied as a tuple i.e. - (4,10) tag from position 4 to 10 (default None)
+        :param data: A dictionary of data for the given tag
+        :param tag_uuid: A UUID used to tie tags in order to demonstrate they're related and form a single concept.  
+                        For example, if tagging the two words "Wells" and "Fargo" as an ORGANIZATION, the tag on both words should have the 
+                        same tag_uuid in order to indicate they are both needed to form the single ORGANIZATION.  If a tag_uuid is provided, it is used
+                        on all tags created in this method.  This may result in multiple nodes or multiple feature values having the same tag_uuid.
+                        For example, if the selector provided results in more than one node being selected, each node would be tagged with the same tag_uuid.  
+                        The same holds true if a content_re value is provided, node_only is set to False, and multiple matches are found for the content_re 
+                        pattern.  In that case, each feature value would share the same UUID.
+                        If no tag_uuid is provided, a new uuid is generated for each tag instance. 
         """
 
         if use_all_content and node_only is None:
@@ -648,7 +710,14 @@ class ContentNode(object):
         elif node_only is None:
             node_only = False
 
-        def tag_node_position(node_to_check, start, end, node_data):
+        def get_tag_uuid(tag_uuid):
+            if tag_uuid:
+                return tag_uuid
+            else:
+                return str(uuid.uuid4())
+
+
+        def tag_node_position(node_to_check, start, end, node_data, tag_uuid):
 
             content_length = 0
 
@@ -659,14 +728,14 @@ class ContentNode(object):
                         node_to_check.add_feature('tag', tag_to_apply,
                                                   Tag(start, end,
                                                       node_to_check.content[start:end],
-                                                      data=node_data))
+                                                      data=node_data, uuid=tag_uuid))
                         return -1
                     elif start < len(node_to_check.content) <= end:
                         node_to_check.add_feature('tag', tag_to_apply,
                                                   Tag(start,
                                                       len(node_to_check.content),
                                                       value=node_to_check.content[start:],
-                                                      data=node_data))
+                                                      data=node_data, uuid=tag_uuid))
 
                 end = end - len(node_to_check.content) + len(separator)
                 content_length = len(node_to_check.content) + len(separator)
@@ -674,7 +743,7 @@ class ContentNode(object):
                     node_to_check.content) - len(separator)
 
             for child_node in node_to_check.children:
-                result = tag_node_position(child_node, start, end, node_data)
+                result = tag_node_position(child_node, start, end, node_data, tag_uuid)
                 content_length = content_length + result
                 if result < 0:
                     return -1
@@ -689,11 +758,11 @@ class ContentNode(object):
 
         for node in self.select(selector):
             if fixed_position:
-                tag_node_position(node, fixed_position[0], fixed_position[1], data)
+                tag_node_position(node, fixed_position[0], fixed_position[1], data, get_tag_uuid(tag_uuid))
 
             else:
                 if not content_re:
-                    node.add_feature('tag', tag_to_apply, Tag(data=data))
+                    node.add_feature('tag', tag_to_apply, Tag(data=data, uuid=get_tag_uuid(tag_uuid)))
                 else:
                     if not use_all_content:
                         if node.content:
@@ -704,24 +773,19 @@ class ContentNode(object):
                         content = node.get_all_content(separator=separator)
 
                     if content is not None:
-                        match = pattern.match(content)
-                        if match:
-                            if node_only:
-                                node.add_feature('tag', tag_to_apply, Tag(data=data))
-                            else:
-                                for index, m in enumerate(match.groups()):
-                                    idx = index + 1
+                        matches = pattern.finditer(content)
 
-                                    if node_only:
-                                        node.add_feature('tag', tag_to_apply,
-                                                         Tag(match.start(idx), match.end(idx), match.group(idx),
-                                                             data=data))
-                                    else:
-                                        # We need to work out where the content is in the child nodes
-                                        start_offset = match.start(idx)
-                                        end_offset = match.end(idx)
+                        if node_only:
+                            # If we are only tagging the node we
+                            # simply need to know if there are any matches
+                            if any(True for _ in matches):
+                                node.add_feature('tag', tag_to_apply, Tag(data=data, uuid=get_tag_uuid(tag_uuid)))
+                        else:
+                            for match in matches:
+                                start_offset = match.span()[0]
+                                end_offset = match.span()[1]
+                                tag_node_position(node, start_offset, end_offset, data, get_tag_uuid(tag_uuid))
 
-                                        tag_node_position(node, start_offset, end_offset, data)
 
     def get_tags(self):
         """
@@ -1264,7 +1328,7 @@ class Document(object):
 
     def remove_label(self, label: str):
         """
-        Remove a label to the document
+        Remove a label from the document
 
         :param label:str Label to remove
         :return: the document
@@ -1273,9 +1337,23 @@ class Document(object):
         return self
 
     @classmethod
-    def from_text(cls, text):
+    def from_text(cls, text, separator=None):
+        """
+        Creates a new Document from the text provided.
+
+        :param text:str  Text to be used as content on the Document's ContentNode(s)
+        :param separator:str   If provided, this string will be used to split the text and the resulting text will be placed on children of the root ContentNode.
+        :return: the document
+        """
         new_document = Document()
-        new_document.content_node = new_document.create_node(node_type='text', content=text)
+        new_document.content_node = new_document.create_node(node_type='text')
+        if text:
+            if separator:
+                for s in text.split(separator):
+                    new_document.content_node.add_child(new_document.create_node(node_type='text', content=s))
+            else:
+                new_document.content_node.content = text
+
         new_document.add_mixin('text')
         return new_document
 
@@ -1440,7 +1518,8 @@ class Document(object):
         """
         registry.add_mixin_to_document(mixin, self)
 
-    def create_node(self, node_type: str, content: str = None, virtual: bool = False, parent: ContentNode = None,
+    def create_node(self, node_type: str, content: Optional[str] = None, virtual: bool = False,
+                    parent: ContentNode = None,
                     index: int = 0):
         """
         Creates a new node for the document.  The new node is not added to the document, but any mixins that have been
@@ -1475,7 +1554,7 @@ class Document(object):
         return content_node
 
     @classmethod
-    def from_file(cls, file):
+    def from_file(cls, file, unpack: bool = False):
         """
         Creates a Document that has a 'file-handle' connector to the specified file.
 
@@ -1484,13 +1563,16 @@ class Document(object):
         :return: A Document connected to the specified file.
         :rtype: Document
         """
-        file_document = Document()
-        file_document.metadata.connector = 'file-handle'
-        file_document.metadata.connector_options.file = file
-        file_document.source.connector = 'file-handle'
-        file_document.source.original_filename = os.path.basename(file)
-        file_document.source.original_path = file
-        return file_document
+        if unpack:
+            Document.from_kdxa(file)
+        else:
+            file_document = Document()
+            file_document.metadata.connector = 'file-handle'
+            file_document.metadata.connector_options.file = file
+            file_document.source.connector = 'file-handle'
+            file_document.source.original_filename = os.path.basename(file)
+            file_document.source.original_path = file
+            return file_document
 
     @classmethod
     def from_url(cls, url, headers=None):
@@ -1641,57 +1723,3 @@ class ModelStore:
 
     def put(self, path: str, content):
         pass
-
-
-class LocalModelStore(ModelStore):
-
-    def __init__(self, store_path: str, force_initialize=False):
-        self.store_path = store_path
-        path = Path(store_path)
-
-        if force_initialize and path.exists():
-            shutil.rmtree(store_path)
-
-        if path.is_file():
-            raise Exception("Unable to load store, since it is pointing to a file?")
-        elif not path.exists():
-            path.mkdir(parents=True)
-
-    def to_dict(self):
-        return {
-            "type": "MODEL",
-            "data": {
-                "path": self.store_path
-            }
-        }
-
-    def get(self, object_path: str):
-        if Path(os.path.join(self.store_path, object_path)).is_file():
-            return open(os.path.join(self.store_path, object_path), 'rb')
-        else:
-            return None
-
-    def put(self, object_path: str, content):
-        path = Path(object_path)
-        with open(os.path.join(self.store_path, path), 'wb') as object_file:
-            object_file.write(content)
-
-
-class RemoteModelStore(ModelStore, RemoteStore):
-
-    def to_dict(self):
-        return {
-            "type": "MODEL",
-            "ref": self.ref
-        }
-
-    def __init__(self, ref: str):
-        self.ref = ref
-
-    def get(self, object_path: str):
-        # TODO implement
-        pass
-
-    def put(self, object_path: str, content):
-        # TODO implement
-        pass
