@@ -11,12 +11,14 @@ class NodeTagger:
     and also just tag the node (ignoring the matching groups)
     """
 
-    def __init__(self, selector, tag_to_apply, content_re=".*", use_all_content=True, node_only=False):
+    def __init__(self, selector, tag_to_apply, content_re=".*", use_all_content=True, node_only=False,
+                 node_tag_uuid=None):
         self.selector = selector
         self.content_re = content_re
         self.use_all_content = use_all_content
         self.tag_to_apply = tag_to_apply
         self.node_only = node_only
+        self.node_tag_uuid = node_tag_uuid
 
     def get_name(self):
         return f"Node Tagger [selector='{self.selector}' use_all_content='{self.use_all_content}']"
@@ -24,11 +26,30 @@ class NodeTagger:
     def process(self, document):
         document.content_node.tag(selector=self.selector, tag_to_apply=self.tag_to_apply, content_re=self.content_re,
                                   use_all_content=self.use_all_content,
-                                  node_only=self.node_only)
+                                  node_only=self.node_only, tag_uuid=self.node_tag_uuid)
 
         return document
 
 
+class NodeTagCopy:
+    """
+    The NodeTagCopy action allows you select nodes specified by the selector and create copies of the existing_tag (if it exists) with the new_tag_name.
+    If a tag with the 'existing_tag_name' does not exist on a selected node, no action is taken for that node.
+    """
+
+    def __init__(self, selector, existing_tag_name, new_tag_name):
+        self.selector = selector
+        self.existing_tag_name = existing_tag_name
+        self.new_tag_name = new_tag_name
+        
+    def get_name(self):
+        return f"Node Tag Copy [selector='{self.selector}' existing_tag_name='{self.existing_tag_name} new_tag_name={self.new_tag_name}']"
+
+    def process(self, document):
+        document.content_node.copy_tag(selector=self.selector, existing_tag_name=self.existing_tag_name, new_tag_name=self.new_tag_name)
+        return document
+
+
 class TextParser:
     """
     The text parser can load a source file as a text document and creates a single content node with the
