@@ -9,8 +9,7 @@ from typing import List, Dict, Optional
 
 import requests
 
-from kodexa.model import Document, Store, DocumentStore
-from kodexa.model.model import RemoteStore, LocalModelStore, RemoteModelStore
+from kodexa.model import Document, Store, DocumentStore, RemoteStore, ModelStore
 
 logger = logging.getLogger('kodexa.stores')
 
@@ -145,6 +144,39 @@ class JsonDocumentStore(Store):
         self.index = 0
 
 
+class RemoteTableDataStore(RemoteStore):
+
+    def __init__(self, ref: str, columns: List[str] = []):
+        self.ref = ref
+        self.columns = columns
+
+    def get_ref(self):
+        return self.ref
+
+    def add(self, row):
+        from kodexa import KodexaPlatform
+
+        url = f"{KodexaPlatform.get_url()}/api/stores/{self.ref.replace(':', '/')}/rows"
+        logger.debug(f"Uploading rows to store {url}")
+
+        row_dict = {}
+        for idx, row_value in enumerate(row):
+            if len(self.columns) == 0 or len(self.columns) <= idx:
+                row_dict[f'col{idx}'] = row_value
+            else:
+                row_dict[self.columns[idx]] = row_value
+
+        doc = requests.post(
+            url,
+            [{'data': row_dict}],
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+        if doc.status_code == 200:
+            return
+        else:
+            logger.error("Unable to post rows to remote store [" + doc.text + "], response " + str(doc.status_code))
+            raise Exception("Unable to post rows to remote store [" + doc.text + "], response " + str(doc.status_code))
+
+
 class TableDataStore(Store):
     """
     Stores data as a list of lists that can represent a table.
@@ -245,6 +277,35 @@ class TableDataStore(Store):
         return TableDataStore(columns=store_dict['data']['columns'], rows=store_dict['data']['rows'])
 
 
+class RemoteDictDataStore(RemoteStore):
+
+    def __init__(self, ref: str):
+        self.ref: str = ref
+
+    def get_ref(self):
+        return self.ref
+
+    def add(self, dict):
+        """
+        Writes a dict to the Data Store
+
+        :param dict: the dict to add to the store
+        """
+        from kodexa import KodexaPlatform
+
+        url = f"{KodexaPlatform.get_url()}/api/stores/{self.ref.replace(':', '/')}/dictionaries"
+        logger.debug(f"Uploading dictionaries to store {url}")
+        doc = requests.post(
+            url,
+            [{'data': dict}],
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+        if doc.status_code == 200:
+            return
+        else:
+            logger.error("Unable to post dict to remote store [" + doc.text + "], response " + str(doc.status_code))
+            raise Exception("Unable to post dict to remote store [" + doc.text + "], response " + str(doc.status_code))
+
+
 class DictDataStore:
     """
     Stores data as a list of dictionaries that can be any structure
@@ -466,8 +527,21 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
         self.query_string: str = query
 
     def get_name(self):
+        """
+        The name of the connector
+
+        :return: 'document-store'
+        """
         return "document-store"
 
+    def get_ref(self) -> str:
+        """
+        Get the reference to the store on the platform (i.e. kodexa/my-store:1.1.0)
+
+        :return: The reference
+        """
+        return self.ref
+
     def to_dict(self):
         return {
             "type": "DOCUMENT",
@@ -475,6 +549,12 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
         }
 
     def get(self, document_id: str) -> Optional[Document]:
+        """
+        Get the document from the remote store using the document's ID
+
+        :param document_id: the ID of the document
+        :return: Optionally the document, if the document is not found we will return None
+        """
         from kodexa import KodexaPlatform
 
         url = f"{KodexaPlatform.get_url()}/api/stores/{self.ref.replace(':', '/')}/contents/{document_id}/content"
@@ -491,6 +571,12 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
             raise Exception("Get document failed [" + doc.text + "], response " + str(doc.status_code))
 
     def get_source(self, document_id: str):
+        """
+        Gets the source for a given document ID from the document store, as per the connector protocol
+
+        :param document_id: document id for which you what the source
+        :return: either the source content (byte-array) or None if the document doesn't exist
+        """
         from kodexa import KodexaPlatform
 
         url = f"{KodexaPlatform.get_url()}/api/stores/{self.ref.replace(':', '/')}/contents/{document_id}/content"
@@ -509,6 +595,11 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
             raise Exception("Get document failed [" + doc.text + "], response " + str(doc.status_code))
 
     def delete(self, document_id: str):
+        """
+        Delete the document with this document ID from the remote document store
+
+        :param document_id: the ID of the document to delete
+        """
         from kodexa import KodexaPlatform
         delete_document_response = requests.delete(
             f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents/{document_id}",
@@ -522,17 +613,29 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
             raise Exception("Delete document failed [" + delete_document_response.text + "], response " + str(
                 delete_document_response.status_code))
 
-    def get_ref(self) -> str:
-        return self.ref
-
     def get_by_path(self, path) -> Optional[Document]:
+        """
+        Return the latest representation document at a given path
+
+        :param path: the path (i.e. /my-folder/file.pdf)
+        :return: The latest document representation for that path
+        """
         hits = self.query_objects(f"path:{path}")
         if len(hits) == 1:
             return self.get(hits[0]['id'])
         else:
             return None
 
-    def put(self, path: str, document: Document):
+    def put(self, path: str, document: Document, force_replace=False):
+        """
+        Put the document into the document store at the given path, note that if there is already a document
+        at this path then the lineage UUID for this document must related to one of the documents the documents
+        that are held at that path
+
+        :param path: the path to the document to use
+        :param document: the document
+        :param force_replace: this will remove all the representations of the document at this path and replace it with this one
+        """
         from kodexa import KodexaPlatform
         try:
             import io
@@ -540,7 +643,7 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
             logger.info(f"Putting document with path {path}")
 
             files = {"document": document.to_msgpack()}
-            data = {"path": path}
+            data = {"path": path, "forceReplace": force_replace}
             content_object_response = requests.post(
                 f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
                 headers={"x-access-token": KodexaPlatform.get_access_token()},
@@ -556,7 +659,7 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
                     content_object_response.status_code))
         except JSONDecodeError:
             logger.error(
-                f"Unable to handle response [{content_object_response.text}], response {content_object_response.status_code}")
+                f"Unable to decode the JSON response")
             raise
 
     def get_next_objects(self):
@@ -572,11 +675,24 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
         else:
             self.objects = content_objects_response.json()['content']
 
-    def query_objects(self, query: str) -> List[Dict]:
+    def query_objects(self, query: str, sort_by=None, sort_direction='asc') -> List[Dict]:
+        """
+        Query the documents in the given document store and a list of the document metadata matches
+
+        :param sort_direction: the sort direction (either asc - ascending or desc - descending)
+        :param sort_by: the name of the metadata field to sort by (ie. createdDate)
+        :param query: A lucene style query for the metadata in the document store
+        :return: A list of dictionaries containing the metadata for the documents
+        """
         from kodexa import KodexaPlatform
+        params = {'query': query}
+        if sort_by is not None:
+            params['sortBy'] = sort_by
+            params['sortDesc'] = sort_direction
+
         list_content = requests.get(
-            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
-            params={'query': query},
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref.replace(':', '/')}/contents",
+            params=params,
             headers={"x-access-token": KodexaPlatform.get_access_token()})
         if list_content.status_code != 200:
             raise Exception(
@@ -584,6 +700,11 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
         return list_content.json()['content']
 
     def list_objects(self) -> List[Dict]:
+        """
+        List the objects in the remote document store
+
+        :return: a list of the dictionaries containing the metadata for the documents in the store
+        """
         from kodexa import KodexaPlatform
         list_content = requests.get(
             f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
@@ -612,7 +733,135 @@ class RemoteDocumentStore(DocumentStore, RemoteStore):
             raise StopIteration
 
 
-class DocumentReference:
+class LocalModelStore(ModelStore):
 
-    def __init__(self, document_store: DocumentStore, path: str):
-        pass
+    def __init__(self, store_path: str, force_initialize=False):
+        self.store_path = store_path
+        path = Path(store_path)
+
+        if force_initialize and path.exists():
+            shutil.rmtree(store_path)
+
+        if path.is_file():
+            raise Exception("Unable to load store, since it is pointing to a file?")
+        elif not path.exists():
+            path.mkdir(parents=True)
+
+    def to_dict(self):
+        return {
+            "type": "MODEL",
+            "data": {
+                "path": self.store_path
+            }
+        }
+
+    def get(self, object_path: str):
+        if Path(os.path.join(self.store_path, object_path)).is_file():
+            return open(os.path.join(self.store_path, object_path), 'rb')
+        else:
+            return None
+
+    def put(self, object_path: str, content):
+        path = Path(object_path)
+        with open(os.path.join(self.store_path, path), 'wb') as object_file:
+            object_file.write(content)
+
+
+class RemoteModelStore(ModelStore, RemoteStore):
+
+    def to_dict(self):
+        return {
+            "type": "MODEL",
+            "ref": self.ref
+        }
+
+    def __init__(self, ref: str):
+        self.ref = ref
+
+    def get_ref(self) -> str:
+        """
+        Get the reference to the store on the platform (i.e. kodexa/my-store:1.1.0)
+
+        :return: The reference
+        """
+        return self.ref
+
+    def delete(self, object_path: str):
+        """
+        Delete the content stored in the model store at the given path
+
+        :param object_path: the path to the content (ie. mymodel.dat)
+        :return: True if deleted, False if there was no file at the path
+        """
+        from kodexa import KodexaPlatform
+        import requests
+        resp = requests.delete(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/fs/{object_path}",
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+
+        if resp.status_code == 200:
+            return True
+        if resp.status_code == 404:
+            return False
+        else:
+            msg = f"Unable to delete model object {resp.text}, status : {resp.status_code}"
+            logger.error(msg)
+            raise Exception(msg)
+
+    def get(self, object_path: str):
+        """
+        Get the bytes for the object at the given path, will return None if there is no object there
+
+        :param object_path: the object path
+        :return: the bytes or None is nothing is at the path
+        """
+        from kodexa import KodexaPlatform
+        import requests
+        resp = requests.get(
+            f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/fs/{object_path}",
+            headers={"x-access-token": KodexaPlatform.get_access_token()})
+
+        if resp.status_code == 200:
+            return resp.content
+        else:
+            msg = f"Unable to get model object {resp.text}, status : {resp.status_code}"
+            logger.error(msg)
+            raise Exception(msg)
+
+    def put(self, path: str, content, force_replace=False):
+        """
+        Put the content into the model store at the given path
+
+        :param path: The path you wish to put the content at
+        :param content: The content for that object
+        :param force_replace: overwrite the existing object if it is there (defaulted to False)
+        """
+        from kodexa import KodexaPlatform
+        import requests
+        try:
+            import io
+
+            files = {"file": content}
+            data = {"path": path, "forceReplace": force_replace}
+            content_object_response = requests.post(
+                f"{KodexaPlatform.get_url()}/api/stores/{self.ref}/contents",
+                headers={"x-access-token": KodexaPlatform.get_access_token()},
+                files=files, data=data)
+
+            if content_object_response.status_code == 200:
+                from addict import Dict
+                content_object = Dict(json.loads(content_object_response.text))
+            elif content_object_response.status_code == 400:
+                from addict import Dict
+                bad_request = Dict(json.loads(content_object_response.text))
+                for error_key in bad_request.errors.keys():
+                    print(bad_request.errors[error_key] + " (" + error_key + ")")
+            else:
+                logger.error("Execution creation failed [" + content_object_response.text + "], response " + str(
+                    content_object_response.status_code))
+                raise Exception("Execution creation failed [" + content_object_response.text + "], response " + str(
+                    content_object_response.status_code))
+        except JSONDecodeError:
+            logger.error(
+                f"Unable to JSON decode the response?")
+            raise
